"""Django application setup."""

from django.apps import AppConfig


class CategoriesConfig(AppConfig):
    """Application configuration for categories."""

    name = "categories"
    verbose_name = "Categories"

    def __init__(self, *args, **kwargs):
        super(CategoriesConfig, self).__init__(*args, **kwargs)
        from django.db.models.signals import class_prepared

        class_prepared.connect(handle_class_prepared)

    def ready(self):
        """Migrate the app after it is ready."""
        from django.db.models.signals import post_migrate

        from .migration import migrate_app

        post_migrate.connect(migrate_app)


def handle_class_prepared(sender, **kwargs):
    """
    See if this class needs registering of fields.
    """
    from .registration import registry
    from .settings import FK_REGISTRY, M2M_REGISTRY

    sender_app = sender._meta.app_label
    sender_name = sender._meta.model_name

    for key, val in list(FK_REGISTRY.items()):
        app_name, model_name = key.split(".")
        if app_name == sender_app and sender_name == model_name:
            registry.register_model(app_name, sender, "ForeignKey", val)

    for key, val in list(M2M_REGISTRY.items()):
        app_name, model_name = key.split(".")
        if app_name == sender_app and sender_name == model_name:
            registry.register_model(app_name, sender, "ManyToManyField", val)



"""This module contains utility functions that are used across the project."""

from django.utils.text import slugify as django_slugify


def slugify(text):
    """Slugify a string. This function is a wrapper to unify the slugify function across the project."""
    return django_slugify(text, allow_unicode=True)



"""
Sphinx configuration.
"""

import os
import sys
from datetime import date
from pathlib import Path

project_root = Path("..").resolve()
sys.path.insert(0, str(project_root / "example"))
sys.path.insert(0, str(project_root))
os.environ["DJANGO_SETTINGS_MODULE"] = "example.settings"

# Setup Django
import django  # NOQA

django.setup()

import categories  # noqa
import categories.urls  # noqa

project = "Django Categories"
copyright = f"2010-{date.today():%Y}, Corey Oordt"

version = categories.__version__
release = categories.__version__

# -- General configuration -----------------------------------------------------

extensions = [
    "myst_parser",
    "sphinx.ext.autodoc",
    "sphinx.ext.viewcode",
    "sphinx.ext.autosummary",
    "sphinx.ext.intersphinx",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.napoleon",
    "sphinx_autodoc_typehints",
    "sphinx.ext.coverage",
    "sphinx.ext.githubpages",
    "sphinxcontrib_django2",
]
autosectionlabel_prefix_document = True
autosectionlabel_maxdepth = 2
autosummary_generate = True
napoleon_attr_annotations = True
napoleon_include_special_with_doc = False
napoleon_include_private_with_doc = True
napoleon_include_init_with_doc = True
myst_enable_extensions = [
    "amsmath",
    "colon_fence",
    "deflist",
    "dollarmath",
    "linkify",
    "replacements",
    "smartquotes",
    "substitution",
    "tasklist",
]
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "django": (
        "https://docs.djangoproject.com/en/stable",
        "https://docs.djangoproject.com/en/stable/_objects",
    ),
}

templates_path = ["_templates"]
source_suffix = [".rst", ".md"]
master_doc = "index"

exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]
pygments_style = "sphinx"
todo_include_todos = False


# -- Options for HTML output ---------------------------------------------------

html_theme = "pydata_sphinx_theme"
html_static_path = ["_static"]
html_css_files = [
    "css/custom.css",
]



"""Manages settings for the categories application."""

import collections

from django.conf import settings
from django.db.models import Q
from django.utils.translation import gettext_lazy as _

DEFAULT_SETTINGS = {
    "ALLOW_SLUG_CHANGE": False,
    "M2M_REGISTRY": {},
    "FK_REGISTRY": {},
    "THUMBNAIL_UPLOAD_PATH": "uploads/categories/thumbnails",
    "THUMBNAIL_STORAGE": settings.DEFAULT_FILE_STORAGE,
    "THUMBNAIL_STORAGE_ALIAS": "default",
    "JAVASCRIPT_URL": getattr(settings, "STATIC_URL", settings.MEDIA_URL) + "js/",
    "SLUG_TRANSLITERATOR": "",
    "REGISTER_ADMIN": True,
    "RELATION_MODELS": [],
}

DEFAULT_SETTINGS.update(getattr(settings, "CATEGORIES_SETTINGS", {}))

if DEFAULT_SETTINGS["SLUG_TRANSLITERATOR"]:
    if isinstance(DEFAULT_SETTINGS["SLUG_TRANSLITERATOR"], collections.Callable):
        pass
    elif isinstance(DEFAULT_SETTINGS["SLUG_TRANSLITERATOR"], str):
        from django.utils.importlib import import_module

        bits = DEFAULT_SETTINGS["SLUG_TRANSLITERATOR"].split(".")
        module = import_module(".".join(bits[:-1]))
        DEFAULT_SETTINGS["SLUG_TRANSLITERATOR"] = getattr(module, bits[-1])
    else:
        from django.core.exceptions import ImproperlyConfigured

        raise ImproperlyConfigured(
            _("%(transliterator) must be a callable or a string.") % {"transliterator": "SLUG_TRANSLITERATOR"}
        )
else:
    DEFAULT_SETTINGS["SLUG_TRANSLITERATOR"] = lambda x: x


# Add all the keys/values to the module's namespace
globals().update(DEFAULT_SETTINGS)

RELATIONS = [Q(app_label=al, model=m) for al, m in [x.split(".") for x in DEFAULT_SETTINGS["RELATION_MODELS"]]]

# The field registry keeps track of the individual fields created.
#  {'app.model.field': Field(**extra_params)}
#  Useful for doing a schema migration
FIELD_REGISTRY = {}

# The model registry keeps track of which models have one or more fields
# registered.
# {'app': [model1, model2]}
# Useful for admin alteration
MODEL_REGISTRY = {}



# Test adding 1 fk string
# Test adding 1 fk dict
# test adding many-to-many
# test adding 1 fk, 1 m2m
import django
from django.test import TestCase

from categories.registration import _process_registry, registry


class CategoryRegistrationTest(TestCase):
    """
    Test various aspects of adding fields to a model.
    """

    def test_foreignkey_string(self):
        FK_REGISTRY = {"flatpages.flatpage": "category"}
        _process_registry(FK_REGISTRY, registry.register_fk)
        from django.contrib.flatpages.models import FlatPage

        self.assertTrue("category" in [f.name for f in FlatPage()._meta.get_fields()])

    def test_foreignkey_dict(self):
        FK_REGISTRY = {"flatpages.flatpage": {"name": "category"}}
        _process_registry(FK_REGISTRY, registry.register_fk)
        from django.contrib.flatpages.models import FlatPage

        self.assertTrue("category" in [f.name for f in FlatPage()._meta.get_fields()])

    def test_foreignkey_list(self):
        FK_REGISTRY = {"flatpages.flatpage": ({"name": "category", "related_name": "cats"},)}
        _process_registry(FK_REGISTRY, registry.register_fk)
        from django.contrib.flatpages.models import FlatPage

        self.assertTrue("category" in [f.name for f in FlatPage()._meta.get_fields()])

    if django.VERSION[1] >= 7:

        def test_new_foreignkey_string(self):
            registry.register_model("flatpages", "flatpage", "ForeignKey", "category")
            from django.contrib.flatpages.models import FlatPage

            self.assertTrue("category" in [f.name for f in FlatPage()._meta.get_fields()])


class Categorym2mTest(TestCase):
    def test_m2m_string(self):
        M2M_REGISTRY = {"flatpages.flatpage": "categories"}
        _process_registry(M2M_REGISTRY, registry.register_m2m)
        from django.contrib.flatpages.models import FlatPage

        self.assertTrue("category" in [f.name for f in FlatPage()._meta.get_fields()])



from categories.base import CategoryBaseAdminForm
from categories.models import Category


class CategoryAdminForm(CategoryBaseAdminForm):
    class Meta:
        model = Category

    def clean_alternate_title(self):
        if self.instance is None or not self.cleaned_data["alternate_title"]:
            return self.cleaned_data["name"]
        else:
            return self.cleaned_data["alternate_title"]



"""View functions for categories."""

from typing import Optional

from django.http import Http404, HttpResponse
from django.shortcuts import get_object_or_404
from django.template.loader import select_template
from django.utils.translation import gettext_lazy as _
from django.views.generic import DetailView, ListView

from .models import Category


def category_detail(
    request, path, template_name="categories/category_detail.html", extra_context: Optional[dict] = None
):
    """Render the detail page for a category."""
    extra_context = extra_context or {}
    path_items = path.strip("/").split("/")
    if len(path_items) >= 2:
        category = get_object_or_404(
            Category, slug__iexact=path_items[-1], level=len(path_items) - 1, parent__slug__iexact=path_items[-2]
        )
    else:
        category = get_object_or_404(Category, slug__iexact=path_items[-1], level=len(path_items) - 1)

    templates = []
    while path_items:
        templates.append("categories/%s.html" % "_".join(path_items))
        path_items.pop()
    templates.append(template_name)

    context = {"category": category}
    if extra_context:
        context.update(extra_context)
    return HttpResponse(select_template(templates).render(context))


def get_category_for_path(path, queryset=Category.objects.all()):
    """Return the category for a path."""
    path_items = path.strip("/").split("/")
    if len(path_items) >= 2:
        queryset = queryset.filter(
            slug__iexact=path_items[-1], level=len(path_items) - 1, parent__slug__iexact=path_items[-2]
        )
    else:
        queryset = queryset.filter(slug__iexact=path_items[-1], level=len(path_items) - 1)
    return queryset.get()


class CategoryDetailView(DetailView):
    """Detail view for a category."""

    model = Category
    path_field = "path"

    def get_object(self, **kwargs):
        """Get the category."""
        if self.path_field not in self.kwargs:
            raise AttributeError(
                "Category detail view %s must be called with " "a %s." % (self.__class__.__name__, self.path_field)
            )
        if self.queryset is None:
            queryset = self.get_queryset()
        try:
            return get_category_for_path(self.kwargs[self.path_field], self.model.objects.all())
        except Category.DoesNotExist:
            raise Http404(
                _("No %(verbose_name)s found matching the query") % {"verbose_name": queryset.model._meta.verbose_name}
            )

    def get_template_names(self):
        """Get the potential template names."""
        names = []
        path_items = self.kwargs[self.path_field].strip("/").split("/")
        while path_items:
            names.append("categories/%s.html" % "_".join(path_items))
            path_items.pop()
        names.extend(super(CategoryDetailView, self).get_template_names())
        return names


class CategoryRelatedDetail(DetailView):
    """Detailed view for a category relation."""

    path_field = "category_path"
    object_name_field = None

    def get_object(self, **kwargs):
        """Get the object to render."""
        if self.path_field not in self.kwargs:
            raise AttributeError(
                "Category detail view %s must be called with " "a %s." % (self.__class__.__name__, self.path_field)
            )
        queryset = super(CategoryRelatedDetail, self).get_queryset()
        try:
            category = get_category_for_path(self.kwargs[self.path_field])
        except Category.DoesNotExist:
            raise Http404(
                _("No %(verbose_name)s found matching the query") % {"verbose_name": queryset.model._meta.verbose_name}
            )
        return queryset.get(category=category)

    def get_template_names(self):
        """Get all template names."""
        names = []
        opts = self.object._meta
        path_items = self.kwargs[self.path_field].strip("/").split("/")
        if self.object_name_field:
            path_items.append(getattr(self.object, self.object_name_field))
        while path_items:
            names.append(
                "%s/category_%s_%s%s.html"
                % (opts.app_label, "_".join(path_items), opts.object_name.lower(), self.template_name_suffix)
            )
            path_items.pop()
        names.append("%s/category_%s%s.html" % (opts.app_label, opts.object_name.lower(), self.template_name_suffix))
        names.extend(super(CategoryRelatedDetail, self).get_template_names())
        return names


class CategoryRelatedList(ListView):
    """List related category items."""

    path_field = "category_path"

    def get_queryset(self):
        """Get the list of items."""
        if self.path_field not in self.kwargs:
            raise AttributeError(
                "Category detail view %s must be called with " "a %s." % (self.__class__.__name__, self.path_field)
            )
        queryset = super(CategoryRelatedList, self).get_queryset()
        category = get_category_for_path(self.kwargs[self.path_field])
        return queryset.filter(category=category)

    def get_template_names(self):
        """Get the template names."""
        names = []
        if hasattr(self.object_list, "model"):
            opts = self.object_list.model._meta
            path_items = self.kwargs[self.path_field].strip("/").split("/")
            while path_items:
                names.append(
                    "%s/category_%s_%s%s.html"
                    % (opts.app_label, "_".join(path_items), opts.object_name.lower(), self.template_name_suffix)
                )
                path_items.pop()
            names.append(
                "%s/category_%s%s.html" % (opts.app_label, opts.object_name.lower(), self.template_name_suffix)
            )
        names.extend(super(CategoryRelatedList, self).get_template_names())
        return names



"""Custom category fields for other models."""

from django.db.models import ForeignKey, ManyToManyField


class CategoryM2MField(ManyToManyField):
    """A many to many field to a Category model."""

    def __init__(self, **kwargs):
        if "to" in kwargs:
            kwargs.pop("to")
        super(CategoryM2MField, self).__init__(to="categories.Category", **kwargs)


class CategoryFKField(ForeignKey):
    """A foreign key to the Category model."""

    def __init__(self, **kwargs):
        if "to" in kwargs:
            kwargs.pop("to")
        super(CategoryFKField, self).__init__(to="categories.Category", **kwargs)



# Generated by Django 3.0.6 on 2020-05-17 18:32

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
        ("categories", "0003_auto_20200306_1050"),
    ]

    operations = [
        migrations.AlterField(
            model_name="categoryrelation",
            name="content_type",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to="contenttypes.ContentType", verbose_name="content type"
            ),
        ),
    ]



from categories.admin import CategoryAdminForm
from categories.base import CategoryBaseAdmin


class CategoryAdmin(CategoryBaseAdmin):
    form = CategoryAdminForm
    list_display = ("name", "alternate_title", "active")
    fieldsets = (
        (None, {"fields": ("parent", "name", "thumbnail", "active")}),
        (
            "Meta Data",
            {
                "fields": ("alternate_title", "alternate_url", "description", "meta_keywords", "meta_extra"),
                "classes": ("collapse",),
            },
        ),
        (
            "Advanced",
            {
                "fields": ("order", "slug"),
                "classes": ("collapse",),
            },
        ),
    )



# Generated by Django 1.10.5 on 2017-02-17 11:11
from __future__ import unicode_literals

import django.db.models.manager
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("categories", "0001_initial"),
    ]

    operations = [
        migrations.AlterModelManagers(
            name="category",
            managers=[
                ("tree", django.db.models.manager.Manager()),
            ],
        ),
        migrations.AlterField(
            model_name="categoryrelation",
            name="relation_type",
            field=models.CharField(
                blank=True,
                help_text="A generic text field to tag a relation, like 'leadphoto'.",
                max_length=200,
                null=True,
                verbose_name="relation type",
            ),
        ),
    ]



"""Alter one or more models' tables with the registered attributes."""

from django.core.management.base import BaseCommand, CommandError


class Command(BaseCommand):
    """
    Alter one or more models' tables with the registered attributes.
    """

    help = "Drop the given field from the given model's table"
    args = "appname modelname fieldname"
    can_import_settings = True
    requires_system_checks = []

    def add_arguments(self, parser):
        """Add app_name, model_name, and field_name arguments to the command."""
        parser.add_argument("app_name")
        parser.add_argument("model_name")
        parser.add_argument("field_name")

    def handle(self, *args, **options):
        """
        Alter the tables.
        """
        from categories.migration import drop_field

        if "app_name" not in options or "model_name" not in options or "field_name" not in options:
            raise CommandError("You must specify an Application name, a Model name and a Field name")

        drop_field(options["app_name"], options["model_name"], options["field_name"])



from categories.models import CategoryBase


class SimpleCategory(CategoryBase):
    """
    A simple of catgorizing example
    """

    class Meta:
        verbose_name_plural = "simple categories"



# Generated by Django 3.0.4 on 2020-03-06 10:50

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
        ("categories", "0002_auto_20170217_1111"),
    ]

    operations = [
        migrations.AlterField(
            model_name="category",
            name="level",
            field=models.PositiveIntegerField(editable=False),
        ),
        migrations.AlterField(
            model_name="category",
            name="lft",
            field=models.PositiveIntegerField(editable=False),
        ),
        migrations.AlterField(
            model_name="category",
            name="rght",
            field=models.PositiveIntegerField(editable=False),
        ),
        migrations.AlterField(
            model_name="categoryrelation",
            name="content_type",
            field=models.ForeignKey(
                limit_choices_to=models.Q(
                    models.Q(("app_label", "simpletext"), ("model", "simpletext")),
                    models.Q(("app_label", "flatpages"), ("model", "flatpage")),
                    _connector="OR",
                ),
                on_delete=django.db.models.deletion.CASCADE,
                to="contenttypes.ContentType",
                verbose_name="content type",
            ),
        ),
    ]



import os

import django
from django.core.files import File
from django.core.files.uploadedfile import UploadedFile
from django.test import TestCase

from categories.models import Category
from example.test_storages import MyTestStorage, MyTestStorageAlias


class TestCategoryThumbnail(TestCase):
    def test_thumbnail(self):
        file_name = "test_image.jpg"
        with open(os.path.join(os.path.dirname(__file__), file_name), "rb") as f:
            test_image = UploadedFile(File(f), content_type="image/jpeg")
            category = Category.objects.create(name="Test Category", slug="test-category", thumbnail=test_image)

            self.assertEqual(category.pk, 1)
            self.assertEqual(category.thumbnail_width, 640)
            self.assertEqual(category.thumbnail_height, 480)

            if django.VERSION >= (4, 2):
                self.assertEqual(category.thumbnail.storage.__class__, MyTestStorageAlias)
            else:
                self.assertEqual(category.thumbnail.storage.__class__, MyTestStorage)



import sys

if sys.version_info >= (3, 0):
    from django_test_migrations.contrib.unittest_case import MigratorTestCase

    class TestMigrations(MigratorTestCase):
        migrate_from = ("categories", "0004_auto_20200517_1832")
        migrate_to = ("categories", "0005_unique_category_slug")

        def prepare(self):
            Category = self.old_state.apps.get_model("categories", "Category")
            Category.tree.create(slug="foo", lft=0, rght=0, tree_id=0, level=0)
            Category.tree.create(slug="foo", lft=0, rght=0, tree_id=0, level=0)
            Category.tree.create(slug="foo", lft=0, rght=0, tree_id=0, level=0)
            for i in range(1, 12):
                Category.tree.create(slug="bar", lft=0, rght=0, tree_id=0, level=0)
            Category.tree.create(slug="baz", lft=0, rght=0, tree_id=0, level=0)
            assert Category.tree.count() == 15

        def test_unique_slug_migration(self):
            Category = self.new_state.apps.get_model("categories", "Category")

            self.assertListEqual(
                list(Category.tree.values_list("slug", flat=True)),
                [
                    "foo",
                    "foo-1",
                    "foo-2",
                    "bar",
                    "bar-01",
                    "bar-02",
                    "bar-03",
                    "bar-04",
                    "bar-05",
                    "bar-06",
                    "bar-07",
                    "bar-08",
                    "bar-09",
                    "bar-10",
                    "baz",
                ],
            )



"""
These functions handle the adding of fields to other models.
"""

from typing import Optional, Type, Union

from collections.abc import Iterable

from django.core.exceptions import FieldDoesNotExist, ImproperlyConfigured
from django.db.models import CASCADE, ForeignKey, ManyToManyField

# from settings import self._field_registry, self._model_registry
from django.utils.translation import gettext_lazy as _

from . import fields

FIELD_TYPES = {
    "ForeignKey": ForeignKey,
    "ManyToManyField": ManyToManyField,
}


class Registry(object):
    """Keeps track of fields and models registered."""

    def __init__(self):
        self._field_registry = {}
        self._model_registry = {}

    def register_model(self, app: str, model_name, field_type: str, field_definitions: Union[str, Iterable]):
        """
        Registration process for Django 1.7+.

        Args:
            app: app name/label
            model_name: name of the model
            field_definitions: a string, tuple or list of field configurations
            field_type: either 'ForeignKey' or 'ManyToManyField'

        Raises:
            ImproperlyConfigured: For incorrect parameter types or missing model.
        """
        from django.apps import apps

        app_label = app

        if isinstance(field_definitions, str):
            field_definitions = [field_definitions]
        elif not isinstance(field_definitions, Iterable):
            raise ImproperlyConfigured(
                _("Field configuration for %(app)s should be a string or iterable") % {"app": app}
            )

        if field_type not in ("ForeignKey", "ManyToManyField"):
            raise ImproperlyConfigured(_('`field_type` must be either `"ForeignKey"` or `"ManyToManyField"`.'))

        try:
            if not hasattr(model_name, "_meta"):
                app_config = apps.get_app_config(app)
                app_label = app_config.label
                model = app_config.get_model(model_name)
            else:
                model = model_name
                model_name = model._meta.model_name
            opts = model._meta
            if app_label not in self._model_registry:
                self._model_registry[app_label] = []
            if model not in self._model_registry[app_label]:
                self._model_registry[app_label].append(model)
        except LookupError:
            raise ImproperlyConfigured(
                'Model "%(model)s" doesn\'t exist in app "%(app)s".' % {"model": model_name, "app": app}
            )

        if not isinstance(field_definitions, (tuple, list)):
            field_definitions = [field_definitions]

        for fld in field_definitions:
            extra_params = {"to": "categories.Category", "blank": True}
            if field_type != "ManyToManyField":
                extra_params["on_delete"] = CASCADE
                extra_params["null"] = True
            if isinstance(fld, str):
                field_name = fld
            elif isinstance(fld, dict):
                if "name" in fld:
                    field_name = fld.pop("name")
                else:
                    continue
                extra_params.update(fld)
            else:
                raise ImproperlyConfigured(
                    _("%(settings)s doesn't recognize the value of %(app)s.%(model)s")
                    % {"settings": "CATEGORY_SETTINGS", "app": app, "model": model_name}
                )
            registry_name = ".".join([app_label, model_name.lower(), field_name])
            if registry_name in self._field_registry:
                continue

            try:
                opts.get_field(field_name)
            except FieldDoesNotExist:
                self._field_registry[registry_name] = FIELD_TYPES[field_type](**extra_params)
                self._field_registry[registry_name].contribute_to_class(model, field_name)

    def register_m2m(self, model, field_name: str = "categories", extra_params: Optional[dict] = None):
        """Register a field name to the model as a many to many field."""
        extra_params = extra_params or {}
        return self._register(model, field_name, extra_params, fields.CategoryM2MField)

    def register_fk(self, model, field_name: str = "category", extra_params: Optional[dict] = None):
        """Register a field name to the model as a foreign key."""
        extra_params = extra_params or {}
        return self._register(model, field_name, extra_params, fields.CategoryFKField)

    def _register(
        self,
        model,
        field_name: str,
        extra_params: Optional[dict] = None,
        field: Type = fields.CategoryFKField,
    ):
        """Does the heavy lifting for registering a field to a model."""
        app_label = model._meta.app_label
        extra_params = extra_params or {}
        registry_name = ".".join((app_label, model.__name__, field_name)).lower()

        if registry_name in self._field_registry:
            return  # raise AlreadyRegistered
        opts = model._meta
        try:
            opts.get_field(field_name)
        except FieldDoesNotExist:
            if app_label not in self._model_registry:
                self._model_registry[app_label] = []
            if model not in self._model_registry[app_label]:
                self._model_registry[app_label].append(model)
            self._field_registry[registry_name] = field(**extra_params)
            self._field_registry[registry_name].contribute_to_class(model, field_name)


registry = Registry()


def _process_registry(registry, call_func):
    """
    Given a dictionary, and a registration function, process the registry.
    """
    from django.apps import apps
    from django.core.exceptions import ImproperlyConfigured

    for key, value in list(registry.items()):
        model = apps.get_model(*key.split("."))
        if model is None:
            raise ImproperlyConfigured(_("%(key)s is not a model") % {"key": key})
        if isinstance(value, (tuple, list)):
            for item in value:
                if isinstance(item, str):
                    call_func(model, item)
                elif isinstance(item, dict):
                    field_name = item.pop("name")
                    call_func(model, field_name, extra_params=item)
                else:
                    raise ImproperlyConfigured(
                        _("%(settings)s doesn't recognize the value of %(key)s")
                        % {"settings": "CATEGORY_SETTINGS", "key": key}
                    )
        elif isinstance(value, str):
            call_func(model, value)
        elif isinstance(value, dict):
            field_name = value.pop("name")
            call_func(model, field_name, extra_params=value)
        else:
            raise ImproperlyConfigured(
                _("%(settings)s doesn't recognize the value of %(key)s")
                % {"settings": "CATEGORY_SETTINGS", "key": key}
            )



"""Admin interface classes."""

from django import forms
from django.contrib import admin
from django.utils.translation import gettext_lazy as _

from .base import CategoryBaseAdmin, CategoryBaseAdminForm
from .genericcollection import GenericCollectionTabularInline
from .models import Category
from .settings import JAVASCRIPT_URL, MODEL_REGISTRY, REGISTER_ADMIN, RELATION_MODELS


class NullTreeNodeChoiceField(forms.ModelChoiceField):
    """A ModelChoiceField for tree nodes."""

    def __init__(self, level_indicator="---", *args, **kwargs):
        self.level_indicator = level_indicator
        super(NullTreeNodeChoiceField, self).__init__(*args, **kwargs)

    def label_from_instance(self, obj) -> str:
        """
        Creates labels which represent the tree level of each node when generating option labels.
        """
        return "%s %s" % (self.level_indicator * getattr(obj, obj._mptt_meta.level_attr), obj)


if RELATION_MODELS:
    from .models import CategoryRelation

    class InlineCategoryRelation(GenericCollectionTabularInline):
        """The inline admin panel for category relations."""

        model = CategoryRelation


class CategoryAdminForm(CategoryBaseAdminForm):
    """The form for a category in the admin."""

    class Meta:
        model = Category
        fields = "__all__"

    def clean_alternate_title(self) -> str:
        """Return either the name or alternate title for the category."""
        if self.instance is None or not self.cleaned_data["alternate_title"]:
            return self.cleaned_data["name"]
        else:
            return self.cleaned_data["alternate_title"]


class CategoryAdmin(CategoryBaseAdmin):
    """Admin for categories."""

    form = CategoryAdminForm
    list_display = ("name", "alternate_title", "active")
    fieldsets = (
        (None, {"fields": ("parent", "name", "thumbnail", "active")}),
        (
            _("Meta Data"),
            {
                "fields": ("alternate_title", "alternate_url", "description", "meta_keywords", "meta_extra"),
                "classes": ("collapse",),
            },
        ),
        (
            _("Advanced"),
            {
                "fields": ("order", "slug"),
                "classes": ("collapse",),
            },
        ),
    )
    autocomplete_fields = ("parent",)

    if RELATION_MODELS:
        inlines = [
            InlineCategoryRelation,
        ]

    class Media:
        js = (JAVASCRIPT_URL + "genericcollections.js",)


if REGISTER_ADMIN:
    admin.site.register(Category, CategoryAdmin)

for model, modeladmin in list(admin.site._registry.items()):
    if model in list(MODEL_REGISTRY.values()) and modeladmin.fieldsets:
        fieldsets = getattr(modeladmin, "fieldsets", ())
        fields = [cat.split(".")[2] for cat in MODEL_REGISTRY if MODEL_REGISTRY[cat] == model]
        # check each field to see if already defined
        for cat in fields:
            for k, v in fieldsets:
                if cat in v["fields"]:
                    fields.remove(cat)
        # if there are any fields left, add them under the categories fieldset
        if len(fields) > 0:
            admin.site.unregister(model)
            admin.site.register(
                model,
                type(
                    "newadmin",
                    (modeladmin.__class__,),
                    {"fieldsets": fieldsets + (("Categories", {"fields": fields}),)},
                ),
            )



from django.test import TestCase

from ..utils import slugify


class TestSlugify(TestCase):
    def test_slugify(self):
        string_dict = {
            "naïve café": "naïve-café",
            "spaced    out": "spaced-out",
            "user@domain.com": "userdomaincom",
            "100% natural": "100-natural",
            "über-cool": "über-cool",
            "façade élégant": "façade-élégant",
            "北京大学": "北京大学",
            "Толстой": "толстой",
            "ñoño": "ñoño",
            "سلام": "سلام",
            "Αθήνα": "αθήνα",
            "こんにちは": "こんにちは",
            "˚č$'\\*>%ˇ'!/": "čˇ",
        }
        for key, value in string_dict.items():
            self.assertEqual(slugify(key), value)



"""Settings management for the editor."""

from django.conf import settings

STATIC_URL = getattr(settings, "STATIC_URL", settings.MEDIA_URL)
if STATIC_URL is None:
    STATIC_URL = settings.MEDIA_URL
MEDIA_PATH = getattr(settings, "EDITOR_MEDIA_PATH", "%seditor/" % STATIC_URL)

TREE_INITIAL_STATE = getattr(settings, "EDITOR_TREE_INITIAL_STATE", "collapsed")

IS_GRAPPELLI_INSTALLED = "grappelli" in settings.INSTALLED_APPS



"""
This is the base class on which to build a hierarchical category-like model.

It provides customizable metadata and its own name space.
"""

from django import forms
from django.contrib import admin
from django.db import models
from django.utils.encoding import force_str
from django.utils.translation import gettext_lazy as _
from mptt.fields import TreeForeignKey
from mptt.managers import TreeManager
from mptt.models import MPTTModel

from .editor.tree_editor import TreeEditor
from .settings import ALLOW_SLUG_CHANGE, SLUG_TRANSLITERATOR
from .utils import slugify


class CategoryManager(models.Manager):
    """
    A manager that adds an "active()" method for all active categories.
    """

    def active(self):
        """
        Only categories that are active.
        """
        return self.get_queryset().filter(active=True)


class CategoryBase(MPTTModel):
    """
    This base model includes the absolute bare-bones fields and methods.

    One could simply subclass this model, do nothing else, and it should work.
    """

    parent = TreeForeignKey(
        "self",
        on_delete=models.CASCADE,
        blank=True,
        null=True,
        related_name="children",
        verbose_name=_("parent"),
    )
    name = models.CharField(max_length=100, verbose_name=_("name"))
    slug = models.SlugField(verbose_name=_("slug"), unique=True)
    active = models.BooleanField(default=True, verbose_name=_("active"))

    objects = CategoryManager()
    tree = TreeManager()

    def save(self, *args, **kwargs):
        """
        Save the category.

        While you can activate an item without activating its descendants,
        It doesn't make sense that you can deactivate an item and have its
        decendants remain active.

        Args:
            args: generic args
            kwargs: generic keyword arguments
        """
        if not self.slug:
            self.slug = slugify(SLUG_TRANSLITERATOR(self.name))[:50]

        super(CategoryBase, self).save(*args, **kwargs)

        if not self.active:
            for item in self.get_descendants():
                if item.active != self.active:
                    item.active = self.active
                    item.save()

    def __str__(self):
        ancestors = self.get_ancestors()
        return " > ".join(
            [force_str(i.name) for i in ancestors]
            + [
                self.name,
            ]
        )

    class Meta:
        abstract = True
        unique_together = ("parent", "name")
        ordering = ("tree_id", "lft")

    class MPTTMeta:
        order_insertion_by = "name"


class CategoryBaseAdminForm(forms.ModelForm):
    """Base admin form for categories."""

    def clean_slug(self):
        """Prune and transliterate the slug."""
        if not self.cleaned_data.get("slug", None) and (self.instance is None or not ALLOW_SLUG_CHANGE):
            self.cleaned_data["slug"] = slugify(SLUG_TRANSLITERATOR(self.cleaned_data["name"]))
        return self.cleaned_data["slug"][:50]

    def clean(self):
        """Clean the data passed from the admin interface."""
        super(CategoryBaseAdminForm, self).clean()

        if not self.is_valid():
            return self.cleaned_data

        opts = self._meta

        # Validate slug is valid in that level
        kwargs = {}
        if self.cleaned_data.get("parent", None) is None:
            kwargs["parent__isnull"] = True
        else:
            kwargs["parent__pk"] = int(self.cleaned_data["parent"].id)
        this_level_slugs = [
            c["slug"] for c in opts.model.objects.filter(**kwargs).values("id", "slug") if c["id"] != self.instance.id
        ]
        if self.cleaned_data["slug"] in this_level_slugs:
            raise forms.ValidationError(_("The slug must be unique among " "the items at its level."))

        # Validate Category Parent
        # Make sure the category doesn't set itself or any of its children as
        # its parent.

        if self.cleaned_data.get("parent", None) is None or self.instance.id is None:
            return self.cleaned_data
        if self.instance.pk:
            decendant_ids = self.instance.get_descendants().values_list("id", flat=True)
        else:
            decendant_ids = []

        if self.cleaned_data["parent"].id == self.instance.id:
            raise forms.ValidationError(_("You can't set the parent of the " "item to itself."))
        elif self.cleaned_data["parent"].id in decendant_ids:
            raise forms.ValidationError(_("You can't set the parent of the " "item to a descendant."))
        return self.cleaned_data


class CategoryBaseAdmin(TreeEditor, admin.ModelAdmin):
    """Base admin class for categories."""

    form = CategoryBaseAdminForm
    list_display = ("name", "active")
    search_fields = ("name",)
    prepopulated_fields = {"slug": ("name",)}

    actions = ["activate", "deactivate"]

    def get_actions(self, request):
        """Get available actions for the admin interface."""
        actions = super(CategoryBaseAdmin, self).get_actions(request)
        if "delete_selected" in actions:
            del actions["delete_selected"]
        return actions

    def deactivate(self, request, queryset):  # NOQA: queryset is not used.
        """
        Set active to False for selected items.
        """
        selected_cats = self.model.objects.filter(pk__in=[int(x) for x in request.POST.getlist("_selected_action")])

        for item in selected_cats:
            if item.active:
                item.active = False
                item.save()
                item.children.all().update(active=False)

    deactivate.short_description = _("Deactivate selected categories and their children")

    def activate(self, request, queryset):  # NOQA: queryset is not used.
        """
        Set active to True for selected items.
        """
        selected_cats = self.model.objects.filter(pk__in=[int(x) for x in request.POST.getlist("_selected_action")])

        for item in selected_cats:
            item.active = True
            item.save()
            item.children.all().update(active=True)

    activate.short_description = _("Activate selected categories and their children")



"""Template tags used to render the tree editor."""

from django.contrib.admin.templatetags.admin_list import _boolean_icon, result_headers
from django.contrib.admin.utils import lookup_field
from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from django.template import Library
from django.utils.encoding import force_str, smart_str
from django.utils.html import conditional_escape, escape, escapejs, format_html
from django.utils.safestring import mark_safe

from categories.editor import settings
from categories.editor.utils import display_for_field

register = Library()

TREE_LIST_RESULTS_TEMPLATE = "admin/editor/tree_list_results.html"
if settings.IS_GRAPPELLI_INSTALLED:
    TREE_LIST_RESULTS_TEMPLATE = "admin/editor/grappelli_tree_list_results.html"


def get_empty_value_display(cl):
    """Get the value to display when empty."""
    if hasattr(cl.model_admin, "get_empty_value_display"):
        return cl.model_admin.get_empty_value_display()

    # Django < 1.9
    from django.contrib.admin.views.main import EMPTY_CHANGELIST_VALUE

    return EMPTY_CHANGELIST_VALUE


def items_for_tree_result(cl, result, form):
    """
    Generates the actual list of data.
    """
    first = True
    pk = cl.lookup_opts.pk.attname
    for field_name in cl.list_display:
        row_class = ""
        try:
            f, attr, value = lookup_field(field_name, result, cl.model_admin)
        except (AttributeError, ObjectDoesNotExist):
            result_repr = get_empty_value_display(cl)
        else:
            if f is None:
                allow_tags = getattr(attr, "allow_tags", False)
                boolean = getattr(attr, "boolean", False)
                if boolean:
                    allow_tags = True
                    result_repr = _boolean_icon(value)
                else:
                    result_repr = smart_str(value)
                # Strip HTML tags in the resulting text, except if the
                # function has an "allow_tags" attribute set to True.
                if not allow_tags:
                    result_repr = escape(result_repr)
                else:
                    result_repr = mark_safe(result_repr)
            else:
                if value is None:
                    result_repr = get_empty_value_display(cl)
                if hasattr(f, "rel") and isinstance(f.rel, models.ManyToOneRel):
                    result_repr = escape(getattr(result, f.name))
                else:
                    result_repr = display_for_field(value, f, "")
                if isinstance(f, models.DateField) or isinstance(f, models.TimeField):
                    row_class = ' class="nowrap"'

        if force_str(result_repr) == "":
            result_repr = mark_safe("&nbsp;")
        # If list_display_links not defined, add the link tag to the first field
        if (first and not cl.list_display_links) or field_name in cl.list_display_links:
            table_tag = {True: "th", False: "td"}[first]

            url = cl.url_for_result(result)
            # Convert the pk to something that can be used in Javascript.
            # Problem cases are long ints (23L) and non-ASCII strings.
            if cl.to_field:
                attr = str(cl.to_field)
            else:
                attr = pk
            value = result.serializable_value(attr)
            result_id = repr(force_str(value))[1:]
            first = False
            result_id = escapejs(value)
            yield mark_safe(
                format_html(
                    smart_str('<{}{}><a href="{}"{}>{}</a></{}>'),
                    table_tag,
                    row_class,
                    url,
                    (
                        format_html(
                            ' onclick="opener.dismissRelatedLookupPopup(window, ' '&#39;{}&#39;); return false;"',
                            result_id,
                        )
                        if cl.is_popup
                        else ""
                    ),
                    result_repr,
                    table_tag,
                )
            )

        else:
            # By default the fields come from ModelAdmin.list_editable, but if we pull
            # the fields out of the form instead of list_editable custom admins
            # can provide fields on a per request basis
            if form and field_name in form.fields:
                bf = form[field_name]
                result_repr = mark_safe(force_str(bf.errors) + force_str(bf))
            else:
                result_repr = conditional_escape(result_repr)
            yield mark_safe(smart_str("<td%s>%s</td>" % (row_class, result_repr)))
    if form and not form[cl.model._meta.pk.name].is_hidden:
        yield mark_safe(smart_str("<td>%s</td>" % force_str(form[cl.model._meta.pk.name])))


class TreeList(list):
    """A list subclass for tree result."""

    pass


def tree_results(cl):
    """Generates a list of results for the tree."""
    if cl.formset:
        for res, form in zip(cl.result_list, cl.formset.forms):
            result = TreeList(items_for_tree_result(cl, res, form))
            if hasattr(res, "pk"):
                result.pk = res.pk
                if res.parent:
                    result.parent_pk = res.parent.pk
                else:
                    res.parent_pk = None
            yield result
    else:
        for res in cl.result_list:
            result = TreeList(items_for_tree_result(cl, res, None))
            if hasattr(res, "pk"):
                result.pk = res.pk
                if res.parent:
                    result.parent_pk = res.parent.pk
                else:
                    res.parent_pk = None
            yield result


def result_tree_list(cl):
    """
    Displays the headers and data list together.
    """
    result = {"cl": cl, "result_headers": list(result_headers(cl)), "results": list(tree_results(cl))}
    return result


result_tree_list = register.inclusion_tag(TREE_LIST_RESULTS_TEMPLATE)(result_tree_list)



"""Category models."""

from functools import reduce

from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.core.files.images import get_image_dimensions
from django.db import models
from django.urls import reverse
from django.utils.encoding import force_str
from django.utils.translation import gettext_lazy as _

from .base import CategoryBase
from .settings import (
    RELATION_MODELS,
    RELATIONS,
    THUMBNAIL_STORAGE,
    THUMBNAIL_STORAGE_ALIAS,
    THUMBNAIL_UPLOAD_PATH,
)

# Determine storage method based on Django version
try:  # Django 4.2+
    from django.core.files.storage import storages

    STORAGE = storages[THUMBNAIL_STORAGE_ALIAS]
except ImportError:
    from django.core.files.storage import get_storage_class

    STORAGE = get_storage_class(THUMBNAIL_STORAGE)()


class Category(CategoryBase):
    """A basic category model."""

    thumbnail = models.FileField(
        upload_to=THUMBNAIL_UPLOAD_PATH,
        null=True,
        blank=True,
        storage=STORAGE,
    )
    thumbnail_width = models.IntegerField(blank=True, null=True)
    thumbnail_height = models.IntegerField(blank=True, null=True)
    order = models.IntegerField(default=0)
    alternate_title = models.CharField(
        blank=True, default="", max_length=100, help_text="An alternative title to use on pages with this category."
    )
    alternate_url = models.CharField(
        blank=True,
        max_length=200,
        help_text="An alternative URL to use instead of the one derived from " "the category hierarchy.",
    )
    description = models.TextField(blank=True, null=True)
    meta_keywords = models.CharField(
        blank=True, default="", max_length=255, help_text="Comma-separated keywords for search engines."
    )
    meta_extra = models.TextField(
        blank=True, default="", help_text="(Advanced) Any additional HTML to be placed verbatim " "in the &lt;head&gt;"
    )

    @property
    def short_title(self):
        """Return the name."""
        return self.name

    def get_absolute_url(self):
        """Return a path."""
        from django.urls import NoReverseMatch

        if self.alternate_url:
            return self.alternate_url
        try:
            prefix = reverse("categories_tree_list")
        except NoReverseMatch:
            prefix = "/"
        ancestors = list(self.get_ancestors()) + [
            self,
        ]
        return prefix + "/".join([force_str(i.slug) for i in ancestors]) + "/"

    if RELATION_MODELS:

        def get_related_content_type(self, content_type):
            """
            Get all related items of the specified content type.
            """
            return self.categoryrelation_set.filter(content_type__name=content_type)

        def get_relation_type(self, relation_type):
            """
            Get all relations of the specified relation type.
            """
            return self.categoryrelation_set.filter(relation_type=relation_type)

    def save(self, *args, **kwargs):
        """Save the category."""
        if self.thumbnail:
            width, height = get_image_dimensions(self.thumbnail.file)
        else:
            width, height = None, None

        self.thumbnail_width = width
        self.thumbnail_height = height

        super(Category, self).save(*args, **kwargs)

    class Meta(CategoryBase.Meta):
        verbose_name = _("category")
        verbose_name_plural = _("categories")

    class MPTTMeta:
        order_insertion_by = ("order", "name")


if RELATIONS:
    CATEGORY_RELATION_LIMITS = reduce(lambda x, y: x | y, RELATIONS)
else:
    CATEGORY_RELATION_LIMITS = []


class CategoryRelationManager(models.Manager):
    """Custom access functions for category relations."""

    def get_content_type(self, content_type):
        """
        Get all the items of the given content type related to this item.
        """
        qs = self.get_queryset()
        return qs.filter(content_type__name=content_type)

    def get_relation_type(self, relation_type):
        """
        Get all the items of the given relationship type related to this item.
        """
        qs = self.get_queryset()
        return qs.filter(relation_type=relation_type)


class CategoryRelation(models.Model):
    """Related category item."""

    category = models.ForeignKey(Category, verbose_name=_("category"), on_delete=models.CASCADE)
    content_type = models.ForeignKey(
        ContentType,
        on_delete=models.CASCADE,
        limit_choices_to=CATEGORY_RELATION_LIMITS,
        verbose_name=_("content type"),
    )
    object_id = models.PositiveIntegerField(verbose_name=_("object id"))
    content_object = GenericForeignKey("content_type", "object_id")
    relation_type = models.CharField(
        verbose_name=_("relation type"),
        max_length=200,
        blank=True,
        null=True,
        help_text=_("A generic text field to tag a relation, like 'leadphoto'."),
    )

    objects = CategoryRelationManager()

    def __unicode__(self):
        return "CategoryRelation"



"""Adds and removes category relations on the database."""

from django.apps import apps
from django.db import DatabaseError, connection, transaction
from django.db.utils import OperationalError, ProgrammingError


def table_exists(table_name):
    """
    Check if a table exists in the database.
    """
    pass


def field_exists(app_name, model_name, field_name):
    """
    Does the FK or M2M table exist in the database already?
    """
    model = apps.get_model(app_name, model_name)
    table_name = model._meta.db_table
    cursor = connection.cursor()
    field_info = connection.introspection.get_table_description(cursor, table_name)
    field_names = [f.name for f in field_info]

    # Return True if the many to many table exists
    field = model._meta.get_field(field_name)
    if hasattr(field, "m2m_db_table"):
        m2m_table_name = field.m2m_db_table()
        try:
            m2m_field_info = connection.introspection.get_table_description(cursor, m2m_table_name)
        except DatabaseError:  # Django >= 4.1 throws DatabaseError
            m2m_field_info = []
        if m2m_field_info:
            return True

    return field_name in field_names


def drop_field(app_name, model_name, field_name):
    """
    Drop the given field from the app's model.
    """
    app_config = apps.get_app_config(app_name)
    model = app_config.get_model(model_name)
    field = model._meta.get_field(field_name)
    with connection.schema_editor() as schema_editor:
        schema_editor.remove_field(model, field)


def migrate_app(sender, *args, **kwargs):
    """
    Migrate all models of this app registered.
    """
    from .registration import registry

    if "app_config" not in kwargs:
        return
    app_config = kwargs["app_config"]

    app_name = app_config.label

    fields = [fld for fld in list(registry._field_registry.keys()) if fld.startswith(app_name)]

    sid = transaction.savepoint()
    for fld in fields:
        model_name, field_name = fld.split(".")[1:]
        if field_exists(app_name, model_name, field_name):
            continue
        model = app_config.get_model(model_name)
        try:
            with connection.schema_editor() as schema_editor:
                schema_editor.add_field(model, registry._field_registry[fld])
                if sid:
                    transaction.savepoint_commit(sid)
        # Django 4.1 with sqlite3 has for some reason started throwing OperationalError
        # instead of ProgrammingError, so we need to catch both.
        except (ProgrammingError, OperationalError):
            if sid:
                transaction.savepoint_rollback(sid)
            continue



from categories import models, settings
from categories.base import CategoryBase


class Category(CategoryBase):
    thumbnail = models.FileField(
        upload_to=settings.THUMBNAIL_UPLOAD_PATH,
        null=True,
        blank=True,
        storage=settings.THUMBNAIL_STORAGE,
    )
    thumbnail_width = models.IntegerField(blank=True, null=True)
    thumbnail_height = models.IntegerField(blank=True, null=True)
    order = models.IntegerField(default=0)
    alternate_title = models.CharField(
        blank=True, default="", max_length=100, help_text="An alternative title to use on pages with this category."
    )
    alternate_url = models.CharField(
        blank=True,
        max_length=200,
        help_text="An alternative URL to use instead of the one derived from " "the category hierarchy.",
    )
    description = models.TextField(blank=True, null=True)
    meta_keywords = models.CharField(
        blank=True, default="", max_length=255, help_text="Comma-separated keywords for search engines."
    )
    meta_extra = models.TextField(
        blank=True, default="", help_text="(Advanced) Any additional HTML to be placed verbatim " "in the &lt;head&gt;"
    )



from django.contrib.auth.models import AnonymousUser
from django.http import Http404
from django.test import Client, RequestFactory, TestCase

from categories import views
from categories.models import Category, CategoryRelation


class MyCategoryRelationView(views.CategoryRelatedDetail):
    model = CategoryRelation


class TestCategoryViews(TestCase):
    fixtures = [
        "musicgenres.json",
    ]

    def setUp(self):
        self.client = Client()
        self.factory = RequestFactory()

    def test_category_detail(self):
        cat0 = Category.objects.get(slug="country", level=0)
        cat1 = cat0.children.get(slug="country-pop")
        cat2 = Category.objects.get(slug="urban-cowboy")
        url = cat0.get_absolute_url()
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        url = cat1.get_absolute_url()
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        url = cat2.get_absolute_url()
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        response = self.client.get("%sfoo/" % url)
        self.assertEqual(response.status_code, 404)

    def test_get_category_for_path(self):
        cat0 = Category.objects.get(slug="country", level=0)
        cat1 = cat0.children.get(slug="country-pop")
        cat2 = Category.objects.get(slug="urban-cowboy")

        result = views.get_category_for_path("/country/country-pop/urban-cowboy/")
        self.assertEqual(result, cat2)
        result = views.get_category_for_path("/country/country-pop/")
        self.assertEqual(result, cat1)
        result = views.get_category_for_path("/country/")
        self.assertEqual(result, cat0)

    def test_categorydetailview(self):
        request = self.factory.get("")
        request.user = AnonymousUser()
        self.assertRaises(AttributeError, views.CategoryDetailView.as_view(), request)

        request = self.factory.get("")
        request.user = AnonymousUser()
        response = views.CategoryDetailView.as_view()(request, path="/country/country-pop/urban-cowboy/")
        self.assertEqual(response.status_code, 200)

        request = self.factory.get("")
        request.user = AnonymousUser()
        self.assertRaises(Http404, views.CategoryDetailView.as_view(), request, path="/country/country-pop/foo/")

    def test_categoryrelateddetailview(self):
        from simpletext.models import SimpleText

        stext = SimpleText.objects.create(name="Test", description="test description")
        cat = Category.objects.get(slug="urban-cowboy")
        cat_rel = CategoryRelation.objects.create(category=cat, content_object=stext)  # NOQA
        request = self.factory.get("")
        request.user = AnonymousUser()
        self.assertRaises(AttributeError, MyCategoryRelationView.as_view(), request)

        request = self.factory.get("")
        request.user = AnonymousUser()
        response = MyCategoryRelationView.as_view()(request, category_path="/country/country-pop/urban-cowboy/")
        self.assertEqual(response.status_code, 200)

        request = self.factory.get("")
        request.user = AnonymousUser()
        self.assertRaises(
            Http404, MyCategoryRelationView.as_view(), request, category_path="/country/country-pop/foo/"
        )



"""The setup script."""

from pathlib import Path

from setuptools import setup


def parse_reqs(filepath: str) -> list:
    """
    Parse a file path containing requirements and return a list of requirements.

    Will properly follow ``-r`` and ``--requirements`` links like ``pip``. This
    means nested requirements will be returned as one list.

    Other ``pip``-specific lines are excluded.

    Args:
        filepath: The path to the requirements file

    Returns:
        All the requirements as a list.
    """
    path = Path(filepath)
    reqstr = path.read_text()
    reqs = []
    for line in reqstr.splitlines():
        line = line.strip()
        if line == "":
            continue
        elif not line or line.startswith("#"):
            # comments are lines that start with # only
            continue
        elif line.startswith("-r") or line.startswith("--requirement"):
            _, new_filename = line.split()
            new_file_path = path.parent / new_filename
            reqs.extend(parse_reqs(new_file_path))
        elif line.startswith("-f") or line.startswith("-i") or line.startswith("--"):
            continue
        elif line.startswith("-Z") or line.startswith("--always-unzip"):
            continue
        else:
            reqs.append(line)
    return reqs


requirements = parse_reqs("requirements.txt")

setup(
    install_requires=requirements,
    py_modules=[],
)



import django.core.files.storage
import mptt.fields
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("contenttypes", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="Category",
            fields=[
                ("id", models.AutoField(verbose_name="ID", serialize=False, auto_created=True, primary_key=True)),
                ("name", models.CharField(max_length=100, verbose_name="name")),
                ("slug", models.SlugField(verbose_name="slug")),
                ("active", models.BooleanField(default=True, verbose_name="active")),
                (
                    "thumbnail",
                    models.FileField(
                        storage=django.core.files.storage.FileSystemStorage(),
                        null=True,
                        upload_to="uploads/categories/thumbnails",
                        blank=True,
                    ),
                ),
                ("thumbnail_width", models.IntegerField(null=True, blank=True)),
                ("thumbnail_height", models.IntegerField(null=True, blank=True)),
                ("order", models.IntegerField(default=0)),
                (
                    "alternate_title",
                    models.CharField(
                        default="",
                        help_text="An alternative title to use on pages with this category.",
                        max_length=100,
                        blank=True,
                    ),
                ),
                (
                    "alternate_url",
                    models.CharField(
                        help_text="An alternative URL to use instead of the one derived from the category hierarchy.",
                        max_length=200,
                        blank=True,
                    ),
                ),
                ("description", models.TextField(null=True, blank=True)),
                (
                    "meta_keywords",
                    models.CharField(
                        default="",
                        help_text="Comma-separated keywords for search engines.",
                        max_length=255,
                        blank=True,
                    ),
                ),
                (
                    "meta_extra",
                    models.TextField(
                        default="",
                        help_text="(Advanced) Any additional HTML to be placed verbatim in the &lt;head&gt;",
                        blank=True,
                    ),
                ),
                ("lft", models.PositiveIntegerField(editable=False, db_index=True)),
                ("rght", models.PositiveIntegerField(editable=False, db_index=True)),
                ("tree_id", models.PositiveIntegerField(editable=False, db_index=True)),
                ("level", models.PositiveIntegerField(editable=False, db_index=True)),
                (
                    "parent",
                    mptt.fields.TreeForeignKey(
                        related_name="children",
                        verbose_name="parent",
                        blank=True,
                        to="categories.Category",
                        on_delete=models.CASCADE,
                        null=True,
                    ),
                ),
            ],
            options={
                "ordering": ("tree_id", "lft"),
                "abstract": False,
                "verbose_name": "category",
                "verbose_name_plural": "categories",
            },
        ),
        migrations.CreateModel(
            name="CategoryRelation",
            fields=[
                ("id", models.AutoField(verbose_name="ID", serialize=False, auto_created=True, primary_key=True)),
                ("object_id", models.PositiveIntegerField(verbose_name="object id")),
                (
                    "relation_type",
                    models.CharField(
                        help_text="A generic text field to tag a relation, like 'leadphoto'.",
                        max_length="200",
                        null=True,
                        verbose_name="relation type",
                        blank=True,
                    ),
                ),
                (
                    "category",
                    models.ForeignKey(verbose_name="category", to="categories.Category", on_delete=models.CASCADE),
                ),
                (
                    "content_type",
                    models.ForeignKey(
                        verbose_name="content type", to="contenttypes.ContentType", on_delete=models.CASCADE
                    ),
                ),
            ],
        ),
        migrations.AlterUniqueTogether(
            name="category",
            unique_together=set([("parent", "name")]),
        ),
    ]



"""URL patterns for the categories app."""

from django.urls import path, re_path
from django.views.generic import ListView

from . import views
from .models import Category

categorytree_dict = {"queryset": Category.objects.filter(level=0)}

urlpatterns = (path("", ListView.as_view(**categorytree_dict), name="categories_tree_list"),)

urlpatterns += (re_path(r"^(?P<path>.+)/$", views.category_detail, name="categories_category"),)



from django.contrib import admin

from categories.admin import CategoryBaseAdmin

from .models import SimpleCategory


class SimpleCategoryAdmin(CategoryBaseAdmin):
    pass


admin.site.register(SimpleCategory, SimpleCategoryAdmin)



from django.contrib.auth.models import User
from django.test import Client, TestCase
from django.test.utils import override_settings
from django.urls import reverse
from django.utils.encoding import smart_str

from categories.models import Category


class TestCategoryAdmin(TestCase):
    def setUp(self):
        self.client = Client()
        self.superuser = User.objects.create_superuser("testuser", "testuser@example.com", "password")

    def test_adding_parent_and_child(self):
        self.client.login(username="testuser", password="password")
        url = reverse("admin:categories_category_add")
        data = {
            "parent": "",
            "name": smart_str("Parent Catégory"),
            "thumbnail": "",
            "filename": "",
            "active": "on",
            "alternate_title": "",
            "alternate_url": "",
            "description": "",
            "meta_keywords": "",
            "meta_extra": "",
            "order": 0,
            "slug": "parent",
            "_save": "_save",
            "categoryrelation_set-TOTAL_FORMS": "0",
            "categoryrelation_set-INITIAL_FORMS": "0",
            "categoryrelation_set-MIN_NUM_FORMS": "1000",
            "categoryrelation_set-MAX_NUM_FORMS": "1000",
        }
        resp = self.client.post(url, data=data)
        self.assertEqual(resp.status_code, 302)
        self.assertEqual(1, Category.objects.count())

        # update parent
        data.update({"name": smart_str("Parent Catégory (Changed)")})
        resp = self.client.post(reverse("admin:categories_category_change", args=(1,)), data=data)
        self.assertEqual(resp.status_code, 302)
        self.assertEqual(1, Category.objects.count())

        # add a child
        data.update(
            {
                "parent": "1",
                "name": smart_str("Child Catégory"),
                "slug": smart_str("child-category"),
            }
        )
        resp = self.client.post(url, data=data)
        self.assertEqual(resp.status_code, 302)
        self.assertEqual(2, Category.objects.count())

        # update child
        data.update({"name": "Child (Changed)"})
        resp = self.client.post(reverse("admin:categories_category_change", args=(2,)), data=data)
        self.assertEqual(resp.status_code, 302)
        self.assertEqual(2, Category.objects.count())

        # test the admin list view
        url = reverse("admin:categories_category_changelist")
        resp = self.client.get(url)
        self.assertEqual(resp.status_code, 200)

    @override_settings(RELATION_MODELS=True)
    def test_addview_get(self):
        self.client.force_login(self.superuser)
        url = reverse("admin:categories_category_add")
        resp = self.client.get(url)
        self.assertEqual(resp.status_code, 200)
        self.assertContains(resp, '<tr class="row1 ">')
        self.assertContains(
            resp,
            '<input type="number" name="categoryrelation_set-0-object_id" class="vIntegerField" '
            'min="0" id="id_categoryrelation_set-0-object_id">',
            html=True,
        )



# Generated by Django 2.0.9 on 2018-10-05 13:59

from django.db import migrations, models

from categories.models import Category


def make_slugs_unique(apps, schema_editor):
    duplicates = Category.tree.values("slug").annotate(slug_count=models.Count("slug")).filter(slug_count__gt=1)
    category_objs = []
    for duplicate in duplicates:
        slug = duplicate["slug"]
        categories = Category.tree.filter(slug=slug)
        count = categories.count()
        i = 0
        for category in categories.all():
            if i != 0:
                category.slug = "{}-{}".format(slug, str(i).zfill(len(str(count))))
                category_objs.append(category)
            i += 1
    Category.objects.bulk_update(category_objs, ["slug"])


class Migration(migrations.Migration):
    dependencies = [
        ("categories", "0004_auto_20200517_1832"),
    ]

    operations = [
        migrations.RunPython(make_slugs_unique, reverse_code=migrations.RunPython.noop),
        migrations.AlterField(
            model_name="category",
            name="slug",
            field=models.SlugField(unique=True, verbose_name="slug"),
        ),
    ]
